// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.messageBus;

import com.microsoft.azure.servicebus.Message;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.mockito.junit.MockitoJUnitRunner;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.notification.provider.azure.messageBus.ProcessNotification;
import org.opengroup.osdu.notification.provider.azure.messageBus.extractor.RequestBodyAdapter;
import org.opengroup.osdu.notification.provider.azure.messageBus.thread.ThreadDpsHeaders;
import org.opengroup.osdu.notification.provider.azure.models.NotificationContent;
import org.opengroup.osdu.notification.provider.azure.util.MDCContextMap;
import org.opengroup.osdu.notification.service.NotificationHandler;

import java.util.HashMap;
import java.util.Map;

import static org.junit.Assert.fail;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class ProcessNotificationTest {
    private static final String dataPartitionId = "opendes";
    private static final String correlationId = "908fcf8d-30c5-4c74-a0ae-ab47b48b7a85";
    private static final String notificationData = "[{\"id\":\"opendes:doc:\",\"kind\":\"opendes:at:wellbore:1.0.0\",\"op\":\"create\"},{\"id\":\"opendes:doc:\",\"kind\":\"opendes:at:wellbore:1.0.0\",\"op\":\"create\"}]";
    private static final String subscriptionName = "TestSubscription";
    private static final String notificationId = "Notification-Test-Subscription";
    private HttpResponse response = new HttpResponse();
    private static final Map<String, String> requestAttributes = new HashMap();
    private final String NOT_ACKNOWLEDGE = "message not acknowledged by client";
    private final String EXCEPTION_NOT_THROWN = "Should Throw Exception";
    @InjectMocks
    private ProcessNotification processNotification;

    @Mock
    private NotificationHandler notificationHandler;
    @Mock
    private RequestBodyAdapter requestBodyAdapter;
    @Mock
    private Message message;
    @Mock
    private NotificationContent notificationContent;
    @Mock
    private MDCContextMap mdcContextMap;
    @Spy
    private ThreadDpsHeaders dpsHeaders;

    @Before
    public void init() {
        requestAttributes.put(DpsHeaders.DATA_PARTITION_ID, dataPartitionId);
        requestAttributes.put(DpsHeaders.CORRELATION_ID, correlationId);
        lenient().doNothing().when(dpsHeaders).setThreadContext(dataPartitionId, correlationId, null);
        lenient().when(mdcContextMap.getContextMap(dataPartitionId, correlationId, null)).thenReturn(new HashMap<>());
        lenient().when(dpsHeaders.getHeaders()).thenReturn(requestAttributes);
        when(notificationContent.getExtractAttributes()).thenReturn(requestAttributes);
        when(notificationContent.getNotificationId()).thenReturn(notificationId);
        when(notificationContent.getData()).thenReturn(notificationData);
    }

    @Test
    public void shouldSuccessfullyPerformNotification() throws Exception {
        response.setResponseCode(200);
        when(requestBodyAdapter.extractNotificationContent(message, subscriptionName)).thenReturn(notificationContent);
        when(notificationHandler.notifySubscriber(notificationId, notificationData, requestAttributes)).thenReturn(response);
        processNotification.performNotification(message, subscriptionName);
        verify(notificationHandler, times(1)).notifySubscriber(notificationId, notificationData, requestAttributes);
        verify(requestBodyAdapter, times(1)).extractNotificationContent(message, subscriptionName);
    }

    @Test
    public void shouldThrowExceptionWhenNotifySubscriberFails() throws Exception {
        response.setResponseCode(400);
        when(requestBodyAdapter.extractNotificationContent(message, subscriptionName)).thenReturn(notificationContent);
        when(notificationHandler.notifySubscriber(notificationId, notificationData, requestAttributes)).thenReturn(response);
        try {
            processNotification.performNotification(message, subscriptionName);
            fail(EXCEPTION_NOT_THROWN);
        } catch (Exception e) {
            verify(notificationHandler, times(1)).notifySubscriber(notificationId, notificationData, requestAttributes);
            verify(requestBodyAdapter, times(1)).extractNotificationContent(message, subscriptionName);
            Assert.assertEquals(NOT_ACKNOWLEDGE, e.getMessage());
        }
    }

    @Test
    public void shouldThrowExceptionWhenNotifySubscriberThrowsException() throws Exception {
        response.setResponseCode(400);
        when(requestBodyAdapter.extractNotificationContent(message, subscriptionName)).thenReturn(notificationContent);
        doThrow(new Exception()).when(notificationHandler).notifySubscriber(notificationId, notificationData, requestAttributes);
        try {
            processNotification.performNotification(message, subscriptionName);
            fail(EXCEPTION_NOT_THROWN);
        } catch (Exception e) {
            verify(notificationHandler, times(1)).notifySubscriber(notificationId, notificationData, requestAttributes);
            verify(requestBodyAdapter, times(1)).extractNotificationContent(message, subscriptionName);
            Assert.assertNotNull(e);
        }
    }
}
