/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.http.HttpClient;
import org.opengroup.osdu.core.common.http.HttpRequest;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.Secret;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.notification.auth.factory.AuthFactory;
import org.opengroup.osdu.notification.auth.interfaces.SecretAuth;
import org.opengroup.osdu.notification.provider.gcp.config.OqmConfigurationProperties;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import java.util.Map;

@Component
@ConditionalOnProperty(name = "oqmDriver")
@RequiredArgsConstructor
@Slf4j
public class OqmNotificationHandler {

  private final OqmConfigurationProperties oqmConfigurationProperties;
  private final HttpClient httpClient;
  private final OqmSubscriptionHandler subscriptionHandler;
  private final AuthFactory authFactory;

  public HttpResponse notifySubscriber(String subscriptionId, String pubsubMessage, Map<String, String> headerAttributes) throws Exception {
    String serviceTopic = headerAttributes.get("topic");
    String dataPartitionId = headerAttributes.get(DpsHeaders.DATA_PARTITION_ID);
    String correlationId = headerAttributes.get(DpsHeaders.CORRELATION_ID);

    if (serviceTopic == null || dataPartitionId == null || correlationId == null) {
      log.warn("Service topic: {}, dataPartitionId: {}, correlationId: {}.", serviceTopic, dataPartitionId, correlationId);
      throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Server error", "Missed header attributes");
    }

    Subscription subscription = subscriptionHandler.getSubscription(dataPartitionId, subscriptionId, serviceTopic);
    Secret secret = subscription.getSecret();

    SecretAuth secretAuth = authFactory.getSecretAuth(secret.getSecretType());
    secretAuth.setSecret(secret);
    String pushUrl = secretAuth.getPushUrl(subscription.getPushEndpoint());

    Map<String, String> requestHeader = secretAuth.getRequestHeaders();
    requestHeader.put(DpsHeaders.CONTENT_TYPE, "application/json");
    requestHeader.put(DpsHeaders.CORRELATION_ID, correlationId);
    requestHeader.put(DpsHeaders.DATA_PARTITION_ID, dataPartitionId);

    HttpRequest request = HttpRequest.post()
            .url(pushUrl)
            .headers(requestHeader)
            .body(pubsubMessage)
            .connectionTimeout(oqmConfigurationProperties.getWaitingTime())
            .build();
    HttpResponse response = httpClient.send(request);
    log.debug("Notification handler | Sending notification to Sub ID: `{}` Endpoint: `{}`.", subscriptionId, subscription.getPushEndpoint());
    return response;
  }
}