/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.cache.IRedisCache;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriver;
import org.opengroup.osdu.core.gcp.oqm.model.*;
import org.opengroup.osdu.notification.provider.gcp.config.OqmConfigurationProperties;
import org.opengroup.osdu.notification.provider.gcp.model.ExternalSubscriptions;
import org.opengroup.osdu.notification.provider.gcp.pubsub.receiver.OqmControlTopicReceiver;
import org.opengroup.osdu.notification.provider.gcp.pubsub.receiver.OqmPublishTopicReceiver;
import org.opengroup.osdu.notification.provider.gcp.pubsub.receiver.OqmServiceTopicReceiver;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Runs once on the service start.
 * 1. Provisions OQM service subscribers for topics provided by Register service.
 * 2. Stores the state of third-party subscribers in a cache.
 * 3. Updates the state of third-party subscribers in a cache on every control topic event.
 * 4. Provisions OQM publish subscribers to process notifications to third-party subscribers.
 */
@Slf4j
@Component
@ConditionalOnProperty(name = "oqmDriver")
@RequiredArgsConstructor
public class OqmSubscriberManager {

    public static final String NOTIFICATION_PREFIX = "notification-";
    public static final String SERVICE_SUFFIX = "-service";
    public static final String PUBLISH_SUFFIX = "-publish";

    private final OqmSubscriptionHandler subscriptionHandler;
    private final ITenantFactory tenantInfoFactory;
    private final OqmDriver driver;
    private final OqmConfigurationProperties properties;
    private final IRedisCache<String, ExternalSubscriptions> subscriptionInfoCache;
    private final OqmNotificationHandler notificationHandler;
    @Deprecated
    private final OqmSubscriptionProvider subscriptionProvider;

    @PostConstruct
    void postConstruct() {
        log.debug("OQM | Provisioning STARTED.");
        provisionSubscriptionInfoCache();
        provisionControlTopicSubscribers();
        provisionServiceSubscribers();
        log.debug("OQM | Provisioning COMPLETED.");
    }

    @PreDestroy
    void onPreDestroy() {
        log.debug("OQM | Destroy STARTED.");
        for (TenantInfo tenantInfo : tenantInfoFactory.listTenantInfo()) {
            subscriptionProvider.deleteControlTopicsSubscription(getControlTopicSubscriptionName(tenantInfo));
        }
        log.debug("OQM | Destroy COMPLETED.");
    }

    /**
     * Initial filling a third-party subscription info cache with entities, which have already stored
     * in Register service's database. Cache should contain all necessary information to send notification to the recipient.
     */
    void provisionSubscriptionInfoCache() {
        for (TenantInfo tenantInfo : tenantInfoFactory.listTenantInfo()) {
            String dataPartitionId = tenantInfo.getDataPartitionId();
            subscriptionInfoCache.put(dataPartitionId, ExternalSubscriptions.builder().subscriptions(Collections.emptyList()).build());
            List<Subscription> subscriptionInfos = subscriptionHandler.getAllSubscriptionInfos(tenantInfo.getDataPartitionId());
            List<Subscription> enrichedSubscriptionInfos = subscriptionInfos.stream()
                    .map(subscription -> subscriptionHandler.getSubscription(dataPartitionId, subscription.getNotificationId(), subscription.getTopic()))
                    .collect(Collectors.toList());
            subscriptionInfoCache.put(dataPartitionId, ExternalSubscriptions.builder().subscriptions(enrichedSubscriptionInfos).build());
            log.debug("Subscription info cache PRELOADED for tenant: {}. Size is: {}.", dataPartitionId, enrichedSubscriptionInfos.size());
        }
    }

    void provisionControlTopicSubscribers() {
        String controlTopicName = properties.getRegisterSubscriberControlTopicName();

        for (TenantInfo tenantInfo : tenantInfoFactory.listTenantInfo()) {
            OqmSubscription subscription = subscriptionProvider.getOrCrateSubscription(getControlTopicSubscriptionName(tenantInfo), controlTopicName, tenantInfo);
            registerControlTopicSubscriber(tenantInfo, subscription);
            log.debug("Control topic subscriber REGISTERED for tenant: {}.", tenantInfo.getDataPartitionId());
        }
    }

    void provisionServiceSubscribers() {
        for (TenantInfo tenantInfo : tenantInfoFactory.listTenantInfo()) {
            log.debug("Provision service topics subscribers for tenant: '{}'", tenantInfo.getDataPartitionId());
            subscriptionProvider.createSubscriptions(tenantInfo);
            List<OqmTopic> topics = subscriptionHandler.getTopics(tenantInfo.getDataPartitionId());
            for (OqmSubscription subscription : buildServiceSubscriptions(topics)) {
                registerServiceTopicSubscriber(tenantInfo, subscription);
                log.debug("Subscriber REGISTERED for service subscription {}.", subscription.getName());
            }
            for (OqmSubscription subscription : buildPublishSubscriptions(topics)) {
                registerPublishTopicSubscriber(tenantInfo, subscription);
                log.debug("Subscriber REGISTERED for publish subscription {}.", subscription.getName());
            }
        }
    }

    private void registerControlTopicSubscriber(TenantInfo tenantInfo, OqmSubscription controlTopicSubscription) {
        OqmSubscriber subscriber = OqmSubscriber.builder()
                .subscription(controlTopicSubscription)
                .messageReceiver(new OqmControlTopicReceiver(subscriptionHandler, subscriptionInfoCache))
                .build();
        driver.subscribe(subscriber, getDestination(tenantInfo));
    }

    /**
     * Register service topic subscriber to multiply service messages per each third-party receiver
     * and enrich them with subscriber-id attribute, then publish into publish topic.
     *
     * @param tenantInfo tenant related information.
     * @param serviceSubscription source subscription of incoming messages to handle by subscriber.
     */
    void registerServiceTopicSubscriber(TenantInfo tenantInfo, OqmSubscription serviceSubscription) {
        OqmSubscriber subscriber = OqmSubscriber.builder()
                .subscription(serviceSubscription)
                .messageReceiver(new OqmServiceTopicReceiver(serviceSubscription, driver, subscriptionInfoCache))
                .build();
        driver.subscribe(subscriber, getDestination(tenantInfo));
    }

    /**
     * Sends notification HTTP call to the push-endpoint specified at Subscription to the third-party subscriber.
     * In case of successful call the outgoing message is properly acknowledged.
     * In case of failure notification delivery retry logic proceeds.
     *
     * @param tenantInfo tenant related information.
     * @param subscription source subscription of incoming messages to handle by subscriber.
     */
    private void registerPublishTopicSubscriber(TenantInfo tenantInfo, OqmSubscription subscription) {
        OqmSubscriber subscriber = OqmSubscriber.builder()
                .subscription(subscription)
                .messageReceiver(new OqmPublishTopicReceiver(subscription, notificationHandler))
                .build();
        driver.subscribe(subscriber, getDestination(tenantInfo));
    }

    /**
     * Build all service OqmSubscriptions for topics provided by register service.
     * OqmSubscription is a representation of message broker subscription,
     * which should be provided on infrastructure side.
     *
     * @param topics topics registered on Register service.
     * @return list of OqmSubscriptions.
     */
    private List<OqmSubscription> buildServiceSubscriptions(List<OqmTopic> topics) {
        return topics.stream()
                .map(topic -> OqmSubscription.builder()
                        .name(NOTIFICATION_PREFIX + topic.getName() + SERVICE_SUFFIX)
                        .topic(topic)
                        .subscriberable(true)
                        .build())
                .collect(Collectors.toList());
    }

    /**
     * Build all publish OqmSubscriptions for topics provided by register service.
     * OqmSubscription is a representation of message broker subscription,
     * which should be provided on infrastructure side.
     *
     * @param topics topics registered on Register service.
     * @return list of OqmSubscriptions.
     */
    private List<OqmSubscription> buildPublishSubscriptions(List<OqmTopic> topics) {
        return topics.stream()
                .map(topic -> OqmSubscription.builder()
                        .name(NOTIFICATION_PREFIX + topic.getName() + PUBLISH_SUFFIX)
                        .topic(OqmTopic.builder().name(topic.getName() + PUBLISH_SUFFIX).build())
                        .subscriberable(true)
                        .build())
                .collect(Collectors.toList());
    }

    @Deprecated
    private String getControlTopicSubscriptionName(TenantInfo tenantInfo) {
        return NOTIFICATION_PREFIX + "control-topic-" + tenantInfo.getDataPartitionId();
    }

    private OqmDestination getDestination(TenantInfo tenantInfo) {
        return OqmDestination.builder()
                .partitionId(tenantInfo.getDataPartitionId())
                .build();
    }
}