/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub.receiver;

import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.gcp.oqm.model.OqmAckReplier;
import org.opengroup.osdu.core.gcp.oqm.model.OqmMessage;
import org.opengroup.osdu.core.gcp.oqm.model.OqmMessageReceiver;
import org.opengroup.osdu.core.gcp.oqm.model.OqmSubscription;
import org.opengroup.osdu.notification.provider.gcp.pubsub.OqmNotificationHandler;
import org.opengroup.osdu.notification.provider.gcp.thread.ThreadScopeContextHolder;

import java.util.Map;

@Slf4j
public class OqmPublishTopicReceiver implements OqmMessageReceiver {

    private final OqmSubscription subscription;
    private final OqmNotificationHandler notificationHandler;

    public OqmPublishTopicReceiver(OqmSubscription subscription, OqmNotificationHandler notificationHandler) {
        this.subscription = subscription;
        this.notificationHandler = notificationHandler;
    }

    @Override
    public void receiveMessage(OqmMessage message, OqmAckReplier replier) {
        String publishTopic = subscription.getTopics().get(0).getName();
        log.debug("OQM | `{}` topic | Received message id: `{}`. Attributes: `{}`.", publishTopic, message.getId(), message.getAttributes());

        try {
            handleMessage(message);
        } catch (Exception e) {
            log.error("OQM | `{}` topic | Message not acknowledged by client", publishTopic, e);
            replier.nack();
        } finally {
            ThreadScopeContextHolder.currentThreadScopeAttributes().clear();
        }

        log.debug("OQM | `{}` topic | Message acknowledged by client", publishTopic);
        replier.ack();
    }

    private void handleMessage(OqmMessage oqmMessage) throws Exception {
        String pubsubMessage = oqmMessage.getData();
        Map<String, String> headerAttributes = oqmMessage.getAttributes();
        String subscriptionId = headerAttributes.get("subscription-id");

        HttpResponse response = notificationHandler.notifySubscriber(subscriptionId, pubsubMessage, headerAttributes);

        if (!response.isSuccessCode()) {
            throw new AppException(response.getResponseCode(), "Notification failed", response.getBody());
        }
    }
}