/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.pubsub;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.cache.IRedisCache;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.gcp.oqm.model.OqmAckReplier;
import org.opengroup.osdu.core.gcp.oqm.model.OqmMessage;
import org.opengroup.osdu.notification.provider.gcp.model.ExternalSubscriptions;
import org.opengroup.osdu.notification.provider.gcp.pubsub.receiver.OqmControlTopicReceiver;

import java.util.HashMap;
import java.util.List;
import java.util.stream.Stream;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class OqmControlTopicReceiverTest {

    @InjectMocks
    private OqmControlTopicReceiver sut;
    @Mock
    private OqmAckReplier replier;
    @Mock
    private IRedisCache<String, ExternalSubscriptions> subscriptionInfoCache;
    @Mock
    private OqmSubscriptionHandler subscriptionHandler;
    @Mock
    private Subscription subscription;
    @Mock
    private ExternalSubscriptions externalSubscriptions;
    @Mock
    private List<Subscription> subscriptionInfos;

    @Test
    public void testReceiveMessageSubscriptionCreated() {
        when(subscriptionHandler.getSubscription(any(), any(), any())).thenReturn(subscription);
        when(subscriptionInfoCache.get(null)).thenReturn(externalSubscriptions);
        when(externalSubscriptions.getSubscriptions()).thenReturn(subscriptionInfos);
        OqmMessage createMessage = OqmMessage.builder().data("Subscription Created").attributes(new HashMap<>()).build();
        sut.receiveMessage(createMessage, replier);
        verify(subscriptionInfos, times(1)).add(subscription);
        verify(subscriptionInfoCache, times(1)).put(any(), any());
        verify(replier, times(1)).ack();
    }

    @Test
    public void testReceiveMessageSubscriptionUpdated() {
        when(subscriptionHandler.getSubscription(any(), any(), any())).thenReturn(subscription);
        when(subscriptionInfoCache.get(null)).thenReturn(externalSubscriptions);
        when(externalSubscriptions.getSubscriptions()).thenReturn(subscriptionInfos);
        OqmMessage updatedMessage = OqmMessage.builder().data("Subscription Updated").attributes(new HashMap<>()).build();
        sut.receiveMessage(updatedMessage, replier);
        verify(subscriptionInfos, times(1)).add(subscription);
        verify(subscriptionInfoCache, times(1)).put(any(), any());
        verify(replier, times(1)).ack();
    }

    @Test
    public void testReceiveMessageSubscriptionDeleted() {
        when(subscriptionInfoCache.get(null)).thenReturn(externalSubscriptions);
        when(externalSubscriptions.getSubscriptions()).thenReturn(subscriptionInfos);
        when(subscriptionInfos.stream()).thenReturn(Stream.of(subscription));
        when(subscription.getNotificationId()).thenReturn("4");
        when(subscription.getTopic()).thenReturn("topic1");
        OqmMessage deletedMessage = OqmMessage.builder()
                .data("Subscription Deleted")
                .attributes(new HashMap<String, String>() {{
                    put("subscription-id", "4");
                    put("topic", "topic1");
                }})
                .build();
        sut.receiveMessage(deletedMessage, replier);
        verify(subscriptionInfos, times(1)).remove(subscription);
        verify(subscriptionInfoCache, times(1)).put(any(), any());
        verify(replier, times(1)).ack();
    }

    @Test
    public void testReceiveMessageUnsupported() {
        OqmMessage unsupportedMessage = OqmMessage.builder().data("Unsupported").attributes(new HashMap<>()).build();
        sut.receiveMessage(unsupportedMessage, replier);
        verify(replier, times(1)).nack();
    }
}