/*
  Copyright 2002-2022 Google LLC
  Copyright 2002-2022 EPAM Systems, Inc

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */

package org.opengroup.osdu.notification.util;

import com.nimbusds.oauth2.sdk.AuthorizationGrant;
import com.nimbusds.oauth2.sdk.ClientCredentialsGrant;
import com.nimbusds.oauth2.sdk.ParseException;
import com.nimbusds.oauth2.sdk.Scope;
import com.nimbusds.oauth2.sdk.TokenRequest;
import com.nimbusds.oauth2.sdk.TokenResponse;
import com.nimbusds.oauth2.sdk.auth.ClientAuthentication;
import com.nimbusds.oauth2.sdk.auth.ClientSecretBasic;
import com.nimbusds.oauth2.sdk.auth.Secret;
import com.nimbusds.oauth2.sdk.id.ClientID;
import com.nimbusds.openid.connect.sdk.OIDCTokenResponseParser;
import java.io.IOException;
import java.net.URI;
import java.util.Objects;
import net.minidev.json.JSONObject;
import org.opengroup.osdu.notification.config.OpenIDProviderConfig;

public class OpenIDTokenProvider {

  private static final OpenIDProviderConfig openIDProviderConfig = OpenIDProviderConfig.Instance();
  private static final String ID_TOKEN = "id_token";
  private final AuthorizationGrant clientGrant = new ClientCredentialsGrant();
  private final URI tokenEndpointURI;
  private final Scope scope;
  private final ClientAuthentication opsClientAuthentication;
  private final ClientAuthentication adminClientAuthentication;
  private final ClientAuthentication editorClientAuthentication;
  private final ClientAuthentication noAccessClientAuthentication;

  public OpenIDTokenProvider() {
    this.tokenEndpointURI = openIDProviderConfig.getProviderMetadata().getTokenEndpointURI();
    this.scope = new Scope(openIDProviderConfig.getScopes());
    this.opsClientAuthentication =
        new ClientSecretBasic(
            new ClientID(openIDProviderConfig.getOpsClientId()),
            new Secret(openIDProviderConfig.getOpsClientSecret())
        );
    this.adminClientAuthentication =
        new ClientSecretBasic(
            new ClientID(openIDProviderConfig.getAdminClientId()),
            new Secret(openIDProviderConfig.getAdminClientSecret())
        );
    this.editorClientAuthentication =
        new ClientSecretBasic(
            new ClientID(openIDProviderConfig.getEditorClientId()),
            new Secret(openIDProviderConfig.getEditorClientSecret())
        );
    this.noAccessClientAuthentication =
        new ClientSecretBasic(
            new ClientID(openIDProviderConfig.getNoAccessClientId()),
            new Secret(openIDProviderConfig.getNoAccessClientSecret())
        );
  }

  public String getOpsAccessToken() {
    try {
      TokenRequest request =
          new TokenRequest(this.tokenEndpointURI, this.opsClientAuthentication, this.clientGrant,
              this.scope);
      return requestToken(request);
    } catch (ParseException | IOException e) {
      throw new RuntimeException(
          "Unable get credentials from TEST_DE_OPS_OPENID_PROVIDER_CLIENT_ID variables", e);
    }
  }

  public String getAdminAccessToken() {
    try {
      TokenRequest request =
          new TokenRequest(this.tokenEndpointURI, this.adminClientAuthentication, this.clientGrant,
              this.scope);
      return requestToken(request);
    } catch (ParseException | IOException e) {
      throw new RuntimeException(
          "Unable get credentials from TEST_DE_ADMIN_OPENID_PROVIDER_CLIENT_ID variables", e);
    }
  }

  public String getEditorAccessToken() {
    try {
      TokenRequest request =
          new TokenRequest(this.tokenEndpointURI, this.editorClientAuthentication, this.clientGrant,
              this.scope);
      return requestToken(request);
    } catch (ParseException | IOException e) {
      throw new RuntimeException(
          "Unable get credentials from TEST_DE_EDITOR_OPENID_PROVIDER_CLIENT_ID variables", e);
    }
  }

  public String getNoAccessToken() {
    try {
      TokenRequest request =
          new TokenRequest(this.tokenEndpointURI, this.noAccessClientAuthentication,
              this.clientGrant, this.scope);
      return requestToken(request);
    } catch (ParseException | IOException e) {
      throw new RuntimeException(
          "Unable get credentials from TEST_NO_ACCESS_OPENID_PROVIDER_CLIENT_ID variables", e);
    }
  }

  private String requestToken(TokenRequest tokenRequest) throws ParseException, IOException {

    TokenResponse parse = OIDCTokenResponseParser.parse(tokenRequest.toHTTPRequest().send());

    if (!parse.indicatesSuccess()) {
      throw new RuntimeException("Unable get credentials variables");
    }

    JSONObject jsonObject = parse.toSuccessResponse().toJSONObject();
    String idTokenValue = jsonObject.getAsString(ID_TOKEN);
    if (Objects.isNull(idTokenValue) || idTokenValue.isEmpty()) {
      throw new RuntimeException("Unable get credentials variables");
    }
    return idTokenValue;
  }

}
