/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.api;

import com.google.common.base.Strings;
import com.sun.jersey.api.client.ClientResponse;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.*;
import org.opengroup.osdu.core.common.notification.ISubscriptionService;
import org.opengroup.osdu.core.common.notification.SubscriptionAPIConfig;
import org.opengroup.osdu.core.common.notification.SubscriptionException;
import org.opengroup.osdu.core.common.notification.SubscriptionFactory;
import org.opengroup.osdu.notification.util.GCPTestUtils;
import org.opengroup.osdu.notification.util.ServicesUtils;
import org.opengroup.osdu.notification.util.TestUtils;

import java.util.Base64;
import java.util.HashMap;
import java.util.Map;

import static org.junit.Assert.*;

public class TestPushEndpointGsa {
  public static final String REGISTER_BASE_URL = "REGISTER_BASE_URL";
  public static final String TOPIC_ID = "TOPIC_ID";
  public static final String INTEGRATION_AUDIENCE = "INTEGRATION_AUDIENCE";
  public static final String OSDU_TENANT = "OSDU_TENANT";
  public static final String STORAGE_HOST = "STORAGE_HOST";
  public static final String LEGAL_HOST = "LEGAL_HOST";
  public static final String DE_OPS_TESTER = "DE_OPS_TESTER";
  private String subscriptionId = null;
  private String notificationId = null;
  private ISubscriptionService subscriptionService;
  private static SubscriptionFactory factory;
  private TestUtils testUtils = new GCPTestUtils();
  private final String suffix = String.valueOf(System.currentTimeMillis());
  private String baseRegisterUrl;
  private String topic;
  private String tenant;
  private String integrationAudience;
  private String storageHost;
  private String legalHost;
  private static final String LEGAL_TAG_NAME = "notification-test-gsa";
  private ServicesUtils servicesUtils;

  @After
  public void deleteResource() throws Exception {
    subscriptionService.delete(subscriptionId);
    servicesUtils.deleteStorageRecords(3, suffix);
    servicesUtils.deleteLegalTag(LEGAL_TAG_NAME);
  }

  @Before
  public void createResource() throws Exception {
    baseRegisterUrl = System.getProperty(REGISTER_BASE_URL, System.getenv(REGISTER_BASE_URL));
    topic = System.getProperty(TOPIC_ID, System.getenv(TOPIC_ID));
    integrationAudience = System.getProperty(INTEGRATION_AUDIENCE, System.getenv(INTEGRATION_AUDIENCE));
    tenant = System.getProperty(OSDU_TENANT, System.getenv(OSDU_TENANT));
    if (Strings.isNullOrEmpty(integrationAudience)) {
      integrationAudience = tenant;
    }
    storageHost = System.getProperty(STORAGE_HOST, System.getenv(STORAGE_HOST));
    legalHost = System.getProperty(LEGAL_HOST, System.getenv(LEGAL_HOST));
    servicesUtils = new ServicesUtils(storageHost, legalHost, testUtils, tenant);
    servicesUtils.createLegalTag(LEGAL_TAG_NAME);
    createResourceInPartition(tenant);
  }

  @Test
  public void testPushEndpoint() throws Exception {
    servicesUtils.createStorageRecords(suffix, 3, LEGAL_TAG_NAME);
    Thread.sleep(10000);
    assertNotNull(subscriptionId);
    assertNotNull(notificationId);
    Map<String, String> headers = new HashMap<>();
    headers.put(DpsHeaders.DATA_PARTITION_ID, tenant);
    ClientResponse clientResponse = testUtils.send(baseRegisterUrl, "/test-gc/state", "GET", testUtils.getOpsToken(), null, "",
        headers, false);
    Map<String, Number> response = new HashMap<>();
    response = testUtils.getResult(clientResponse, 200, response.getClass());
    assertNotNull(response);
    assertTrue(response.containsKey(suffix));
    assertTrue(response.get(suffix).longValue() >= 3);
  }

  private void createResourceInPartition(String partitionId) throws Exception {

    SubscriptionAPIConfig config = SubscriptionAPIConfig.builder().rootUrl(baseRegisterUrl).build();
    factory = new SubscriptionFactory(config);

    Map<String, String> headers = new HashMap<>();
    headers.put(DpsHeaders.DATA_PARTITION_ID, partitionId);
    headers.put(DpsHeaders.AUTHORIZATION, testUtils.getOpsToken());
    DpsHeaders dpsHeaders = DpsHeaders.createFromMap(headers);
    subscriptionService = factory.create(dpsHeaders);

    Subscription subscription = new Subscription();
    subscription.setName("subscription-integration-test-gsa-" + suffix);
    subscription.setDescription("subscription created for gsa integration test " + suffix);
    subscription.setTopic(topic);
    subscription.setPushEndpoint(getPushUrl());

    Secret gsaSecret = new GsaSecret();
    GsaSecretValue gsaSecretValue = new GsaSecretValue();
    gsaSecretValue.setAudience(integrationAudience);
    String opsTester = System.getProperty(DE_OPS_TESTER, System.getenv(DE_OPS_TESTER));
    gsaSecretValue.setKey(new String(Base64.getDecoder().decode(opsTester)));
    gsaSecret.setSecretType("GSA");
    ((GsaSecret) gsaSecret).setValue(gsaSecretValue);
    subscription.setSecret(gsaSecret);
    try {
      Subscription subscriptionCreated = subscriptionService.create(subscription);
      notificationId = subscriptionCreated.getNotificationId();
      subscriptionId = subscriptionCreated.getId();
    } catch (SubscriptionException e) {
      System.out.println("Subscription exception inner response : " + e.getHttpResponse());
      throw e;
    }
  }

  private String getPushUrl() {
    return baseRegisterUrl + "/test-gc/gsa-challenge/" + suffix;
  }

}
